/*
 *    MessageRouter.java
 *    Copyright (C) 2002 New Zealand Digital Library, http://www.nzdl.org
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package org.greenstone.gsdl3.core;

import org.greenstone.gsdl3.util.*;
import org.greenstone.gsdl3.service.*;
import org.greenstone.gsdl3.comms.*;
import org.greenstone.gsdl3.collection.*;

// XML classes
import org.w3c.dom.Node; 
import org.w3c.dom.NodeList; 
import org.w3c.dom.Document; 
import org.w3c.dom.Element; 
import org.xml.sax.InputSource;
import javax.xml.parsers.*;

// other java classes
import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.io.Reader;
import java.io.StringReader;

/**
 * The hub of a Greenstone system. 
 *
 * Accepts  XML requests (via process method of ModuleInterface) and routes them to the appropriate collection or 
 * service or external entity.
 *
 * contains a map of module objects - may be services, collections, comms 
 * objects talking to other MessageRouters etc.
 *
 *
 * @author <a href="mailto:kjdon@cs.waikato.ac.nz">Katherine Don</a>
 * @version $Revision: 4000 $
 * @see ModuleInterface
 * @see Collection
 * @see ServiceRack
 * @see Communicator
 *
 */
public class MessageRouter implements  ModuleInterface {

    /** site home - the home directory for the site */
    protected String site_home_=null;
    /** the http address for this site */
    protected String site_http_address_=null;
    
    /** the global name for this site */
    protected String global_site_name_=null;

    /** map of names to Module objects */
    protected HashMap module_map_=null;

    /** container Document to create XML Nodes */
    protected Document doc_=null; 
    /** the full description of this site */
    protected Element description_ = null;

    // should these things be separated into local and remote??

    /** list of collections that can be reached */
    protected Element collection_list_ = null; 
    /** list of service clusters that can be reached */
    protected Element cluster_list_ = null; 
    /** list of single services that can be reached
     * can we have single services? or chould all be part of a cluster */
    protected Element service_list_ = null; 

    /** list of sites that can be reached */
    protected Element site_list_ = null; 

    /** a converter class to parse XML and create Docs */
    protected XMLConverter converter_=null;

    //***************************************************************
    // public methods
    //***************************************************************

    /** constructor */
    public MessageRouter() {
	module_map_ = new HashMap();
	converter_ = new XMLConverter();
	doc_ = converter_.newDOM();
	
	
    }

    /** site_home_ must be set before configure called */
    public void setSiteHome(String home) {
	site_home_ = home;
    }

    /**
     * configures the system
     *
     * looks in site_home_/collect for collections, reads config file
     * site_home_/sitecfg.xml
     *
     */
    public boolean configure() { 

	if (site_home_==null) {
	    System.err.println("You must set site_home_ before calling configure");
	    return false;
	}

    	// start the description_ XML tree
	description_ = doc_.createElement(GSXML.RESPONSE_ELEM);
	description_.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_DESCRIBE);
	description_.setAttribute(GSXML.FROM_ATT, "");
	collection_list_ = doc_.createElement(GSXML.COLLECTION_ELEM+GSXML.LIST_MODIFIER);
	cluster_list_ = doc_.createElement(GSXML.CLUSTER_ELEM+GSXML.LIST_MODIFIER);
	service_list_ = doc_.createElement(GSXML.SERVICE_ELEM+GSXML.LIST_MODIFIER);
	site_list_ = doc_.createElement(GSXML.SITE_ELEM+GSXML.LIST_MODIFIER);


	// read thru own config file - create services and connect to sites
	File configFile = new File(GSFile.siteConfigFile(site_home_));

	if (configFile.exists() ) {
	    configureSystem(configFile);
	}
	else {
	    System.err.println("MessageRouter: site config file: "+configFile.getPath()+" not found!");
	    return false;
	    
	}

	// read thru the collect directory and activate all the valid collections
	File collectDir = new File(GSFile.collectDir(site_home_));
	if (collectDir.exists()) {
	    System.out.println("Reading thru directory "+collectDir.getPath()+" to find collections.");
	    File[] contents = collectDir.listFiles();
	    for (int i=0; i<contents.length;i++) {
		if(contents[i].isDirectory()) {
		    
		    String colName = contents[i].getName();
		    if (!colName.startsWith("CVS")) {
			activateCollection(colName);
		    }
		}
	    }
	} // collectDir
	
	description_.appendChild(collection_list_);

	return true;
	
    }

    /**
     * Process an XML request - as a String
     *
     * @param xml_in the request to process
     * @return the response - contains any error messages
     * @see String
     */
    public String process(String xml_in) {
	
	Document doc = converter_.getDOM(xml_in);
	
	Element result = process(doc.getDocumentElement());
	return converter_.getString(result);
    }
    
    /**
     * Process an XML request - as a DOM Element
     *
     * @param xml_in the message to process - should be <message>
     * @return the response - contains any error messages
     * @see Element
     */
    public Element process(Element message) {

	//System.out.println("MR received request");
	//System.out.println(converter_.getString(message));
			   
	// check that its a correct message tag
	if (!message.getTagName().equals(GSXML.MESSAGE_ELEM)) {
	    System.err.println("Invalid message. GSDL message should start with <"+GSXML.MESSAGE_ELEM+">, instead it starts with:"+message.getTagName()+".");
	    return null;
	}

	NodeList requests = message.getElementsByTagName(GSXML.REQUEST_ELEM);

	Element mainResult = doc_.createElement(GSXML.MESSAGE_ELEM);
	
	// empty request
	if (requests.getLength()==0) {
	    return mainResult;
	}

	Document message_doc = message.getOwnerDocument();

	// for now, just process each request one by one, and append the results to mainResult
	// Note: if you add an element to another node in the same document, it
	// gets removed from where it was. This changes the node list - you cant iterate over the node list in a normal manner if you are moving elements out of it
	int num_requests = requests.getLength();
	for (int i=0; i< num_requests; i++) {
	    Node result=null;
	    Element req = (Element)requests.item(0);
	    String path = req.getAttribute(GSXML.TO_ATT); // returns "" if no att of this name
	    if (path.equals("")) {
		// its a message for the message router
		result = processMessage(req);
		mainResult.appendChild(doc_.importNode(result, true));
	    } else {
		// find the module to pass it on to
		// need to put the request into a message element
		Element mess = message_doc.createElement(GSXML.MESSAGE_ELEM);
		mess.appendChild(req);
		String obj = GSPath.getFirstLink(path);
		if (module_map_.containsKey(obj)) {
		    result = ((ModuleInterface)module_map_.get(obj)).process(mess);
		    if (result !=null ) {
			// append the contents of the message to the mainResult - there will only be one response at this stage
			mainResult.appendChild(doc_.importNode(GSXML.getChildByTagName(result, GSXML.RESPONSE_ELEM), true));
		    } else {
			System.err.println("MessageRouter Error: request had null result!");
		    }
		} else {    
		    System.err.println("MessageRouter Error: request has illegal module name in:\n"+converter_.getString(req));
		}
	    }
	    
	} // for each request
	
	//System.out.println("MR returned response");
	//System.out.println(converter_.getString(mainResult));

	return mainResult;
	
    }
	
    //********************************************************************
    // auxiliary configure methods
    // *******************************************************************

    /**
     * configures the message router based on the site config file 
     * - siteConfig.xml
     *
     */
    protected boolean configureSystem(File file) {

	System.out.println("MessageRouter:configuring site");
	Element config = converter_.getDOM(file).getDocumentElement();

	Element local_site_name = (Element)GSXML.getChildByTagName(config, 
					     GSXML.SITE_NAME_ELEM);
	if (local_site_name == null) {
	    System.err.println("MessageRouter configure error:no site name in config file");
	    return false;
	} else {
	    global_site_name_ = local_site_name.getAttribute(GSXML.VALUE_ATT);
	}
	
	Element http_address = (Element)GSXML.getChildByTagName(config, 
					    GSXML.SITE_HTTP_ADDRESS_ELEM);
	if (http_address == null) {
	    System.err.println("MessageRouter configure error: no http address  in config file");
	    return false;
	} else {
	    site_http_address_ = http_address.getAttribute(GSXML.VALUE_ATT);
	}
	
	// load up the service clusters
	System.out.println("loading service clusters ...");
	Node cluster_list = GSXML.getChildByTagName(config, 
			     GSXML.CLUSTER_ELEM+GSXML.LIST_MODIFIER);
	if (cluster_list == null) {
	    System.out.println("... none to be loaded");
	} else {
	    NodeList service_clusters = cluster_list.getChildNodes();
	    if (service_clusters.getLength()==0) {
		System.out.println("... none to be loaded");
	    } else {
		
		for (int i=0; i<service_clusters.getLength(); i++) {
		    if (service_clusters.item(i).getNodeName().equals(GSXML.CLUSTER_ELEM)) {
			Element cluster = (Element)service_clusters.item(i);
			String name = cluster.getAttribute(GSXML.NAME_ATT);
			System.out.println("..."+name);
			ServiceCluster sc = new ServiceCluster();
			sc.setSiteHome(site_home_);
			sc.setSiteAddress(site_http_address_);
			sc.setClusterName(name);
			sc.setMessageRouter(this);
			sc.configure(cluster);
			module_map_.put(name, sc);
			//add short info to 
			Element e = doc_.createElement(GSXML.CLUSTER_ELEM);
			e.setAttribute(GSXML.NAME_ATT, name);
			cluster_list_.appendChild(e);

		    }
		}
	    }
	    
	}

	// load up the individual services
	System.out.println("loading service modules...");
	
	Node module_list = GSXML.getChildByTagName(config, 
						   GSXML.SERVICE_CLASS_ELEM+GSXML.LIST_MODIFIER);
	if (module_list == null) {
	    System.out.println("... none to be loaded");
	} else {
	    NodeList service_impls = module_list.getChildNodes();
	    if (service_impls.getLength()==0) {
		System.out.println("... none to be loaded");
	    } else {
		
		Element service_request = doc_.createElement(GSXML.REQUEST_ELEM);
		service_request.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_DESCRIBE);
		service_request.setAttribute(GSXML.INFO_ATT, GSXML.SERVICE_ELEM+GSXML.LIST_MODIFIER);
		
		for(int i=0; i<service_impls.getLength(); i++) {
		    if (service_impls.item(i).getNodeName().equals(GSXML.SERVICE_CLASS_ELEM)) {
			Element n = (Element)service_impls.item(i);
			String service_name = n.getAttribute(GSXML.NAME_ATT);
			System.out.println("..."+service_name);
			try {
			    ServiceRack s = (ServiceRack)Class.forName("org.greenstone.gsdl3.service."+service_name).newInstance();
			    s.setSiteHome(site_home_);
			    s.setSiteAddress(site_http_address_);
			    s.setMessageRouter(this);
			    
			    // pass the XML node to the service for service configuration
			    s.configure(n);
			    
			    // find out the supported services for this service module
			    Element service_response = (Element) s.process(service_request);
			    NodeList services = service_response.getElementsByTagName(GSXML.SERVICE_ELEM);
			    if (services.getLength()==0) {
				System.err.println("MessageRouter configure error: servicesImpl "+service_name+" has no services!");
			    } else {
				for (int j=0; j<services.getLength();j++) {
				    String service = ((Element)services.item(j)).getAttribute(GSXML.NAME_ATT);
				    module_map_.put(service, s);
				    
				// add short info to service_list_ XML
				    service_list_.appendChild(doc_.importNode(services.item(j), true));
				}
			    }
			} catch (Exception e ) {
			    System.err.println("MessageRouter configure exception:  in ServiceRack class specification:  "+ e.getMessage());
			    
			} 
		    } // if
		} // for each service module
	    }
	}
	
	// load up the sites
	System.out.println("loading external sites...");
	Node site_list = GSXML.getChildByTagName(config, GSXML.SITE_ELEM+GSXML.LIST_MODIFIER);
	if (site_list ==null ) {
	    System.out.println("...none found");
	} else {
	    NodeList sites = site_list.getChildNodes();
	    if (sites.getLength()==0) {
		System.out.println("...none found");
	    } else {
		
		for (int i=0; i<sites.getLength(); i++) {
		    if (sites.item(i).getNodeName().equals(GSXML.SITE_ELEM)) {
			Element s = (Element)sites.item(i);
			Communicator comm=null;
			String type = s.getAttribute(GSXML.TYPE_ATT);
			String name = s.getAttribute(GSXML.NAME_ATT);
			if (type.equals(GSXML.COMM_TYPE_SOAP_JAVA)) {
			    comm = new SOAPCommunicator();
			    if (comm.configure(s)) {
				comm.setLocalSiteName(global_site_name_);
			    
				// add to map of modules
				module_map_.put(name, comm);
				site_list_.appendChild(doc_.importNode(s, true));
				// need to get collection list and service 
				// list from here- this may introduce a 
				// configure problem - leave as  is for now, 
				// but fix up later - kjdon
				if (!getRemoteSiteInfo(comm, name)) {
				    System.err.println("couldn't get info from site "+name);
				}
			    } else {
				System.err.println("couldn't configure soap site:"+name);
			    }
			    
			} else {
			    System.err.print("cant talk to server of type:"+type);
			    System.err.println(", so not making a connection to "+name);
			}
			
		    }
		}
	    }
	}
	// add children to description_
	description_.appendChild(cluster_list_);
	description_.appendChild(service_list_);
	description_.appendChild(site_list_);
		
	return true;

    }

    
    /** get site info from external site
     *
     * @param comm - the communicator object for the external site
     * @param site_name - the name of the external site
     * @return true if successful
     */
    protected boolean getRemoteSiteInfo(Communicator comm, String site_name) {
    
	System.out.println("MessageRouter: getting info from site:"+site_name);

	Element info_request = doc_.createElement(GSXML.MESSAGE_ELEM);
	Element req = doc_.createElement(GSXML.REQUEST_ELEM);
	info_request.appendChild(req);
	req.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_DESCRIBE);

	// process the message
	Element info_response = comm.process(info_request);
	if (info_response == null) {
	    return false;
	}
	// collection info
	NodeList colls = info_response.getElementsByTagName(GSXML.COLLECTION_ELEM);
	if (colls.getLength()>0) {
	    for (int i=0; i<colls.getLength(); i++) {
		Element e = (Element)colls.item(i);
		String col_name = e.getAttribute(GSXML.NAME_ATT);
		// add the info to own coll list - may want to keep 
		// this separate in future - so can distinguish own and 
		// other collections ??
		e.setAttribute(GSXML.NAME_ATT, GSPath.prependLink(col_name, site_name));
		collection_list_.appendChild(doc_.importNode(e, true));
	    }
	}

	// service info
	NodeList services = info_response.getElementsByTagName(GSXML.SERVICE_ELEM);
	if (services.getLength()>0) {
	    for (int i=0; i<services.getLength(); i++) {
		Element e = (Element)services.item(i);
		String serv_name = e.getAttribute(GSXML.NAME_ATT);
		e.setAttribute(GSXML.NAME_ATT, GSPath.prependLink(serv_name, site_name));
		service_list_.appendChild(doc_.importNode(e, true));
	    }
	}
	
	// serviceCluster info
	NodeList clusters = info_response.getElementsByTagName(GSXML.CLUSTER_ELEM);
	if (clusters.getLength()>0) {
	    for (int i=0; i<clusters.getLength(); i++) {
		Element e = (Element)clusters.item(i);
		String clus_name = e.getAttribute(GSXML.NAME_ATT);
		e.setAttribute(GSXML.NAME_ATT, GSPath.prependLink(clus_name, site_name));
		cluster_list_.appendChild(doc_.importNode(e, true));
	    }
	}
	return true;
    }


    /** creates and configures a new collection
     *
     *@param col_name the name of the collection
     *@return true if collection created ok
     */
    protected boolean activateCollection(String col_name) {

	System.out.println("MessageRouter:Activating collection: "+col_name+".");
			  
	Collection c = new Collection();
	c.setCollectionName(col_name);
	c.setSiteHome(site_home_);
	c.setSiteAddress(site_http_address_);
	if (c.configure()) {
	    // this could be a reactivation, so delete the old version
	    deactivate(GSXML.COLLECTION_ELEM, col_name);
	    // add to list of collections
	    module_map_.put(col_name, c);
	    //add short description_ to collection_list_
	    Element e = doc_.createElement(GSXML.COLLECTION_ELEM);
	    e.setAttribute(GSXML.NAME_ATT, col_name);
	    collection_list_.appendChild(e);
	    return true;
	} else {
	    System.err.println("MessageRouter:Couldn't configure collection: "+
			       col_name+".");
	    return false;
	}
	
    }

    protected boolean activateServiceCluster(String cluster_name) {
	return false;

    }

    protected boolean activateServiceRack(String module_name) {
	return false;
    }

    //*****************************************************************
    // auxiliary process methods
    //*****************************************************************

    /** handles requests made to the MessageRouter itself 
     *
     * @param req - the request Element- <request>
     * @return the result Element - should be <response>
     */
    protected Element processMessage(Element req) {

	// message for self, should be type=describe/configure at this stage
	String type = req.getAttribute(GSXML.TYPE_ATT);
	Element response = doc_.createElement(GSXML.RESPONSE_ELEM);
	response.setAttribute(GSXML.FROM_ATT, "");
	if (type.equals(GSXML.REQUEST_TYPE_DESCRIBE)) {
	    // check the param list
	    Element param_list = (Element) GSXML.getChildByTagName(req, GSXML.PARAM_ELEM+GSXML.LIST_MODIFIER);
	    if (param_list == null) {
		// no params - just return teh full description
		return description_; // this is already in a response elem
	    }
	    System.out.println("poarams found, getting subset");
	    // use the new response with the appropriate info 
	    response.setAttribute(GSXML.TYPE_ATT, GSXML.REQUEST_TYPE_DESCRIBE);
	    
	    // go through the param list and see what components are wanted
	    Element param = (Element) param_list.getFirstChild();
	    while (param != null) {
		if (!param.getNodeName().equals(GSXML.PARAM_ELEM)) {
		    System.err.println("Warning: Non-param in paramList (ignored).");
		    continue;
		}
		// Identify the structure information desired
		if (param.getAttribute(GSXML.NAME_ATT) == GSXML.INFO_PARAM ) {
		    String info = param.getAttribute(GSXML.VALUE_ATT);
		    if (info.equals(GSXML.COLLECTION_ELEM+GSXML.LIST_MODIFIER)) {
			response.appendChild(collection_list_);
			
		    } else if (info.equals(GSXML.CLUSTER_ELEM+GSXML.LIST_MODIFIER)) {
			response.appendChild(cluster_list_);
			
		    } else if (info.equals(GSXML.SERVICE_ELEM+GSXML.LIST_MODIFIER)) {
			return response;
		    } else if (info.equals(GSXML.SITE_ELEM+GSXML.LIST_MODIFIER)) {
			response.appendChild(site_list_);
		    } 
		}
		param = (Element)param.getNextSibling();
	    }
	    return response;
	    
	} else if (type.equals(GSXML.REQUEST_TYPE_CONFIGURE)) {
	    
	    // a list of configure requests - should put any error messages
	    // or success messages into response
	    NodeList commands = req.getElementsByTagName(GSXML.CONFIGURE_ELEM);
	    for (int i=0; i<commands.getLength(); i++) {
		// all the commands should be Elements
		Element elem = (Element)commands.item(i);
		String action = elem.getAttribute(GSXML.ACTION_ATT);
		String module_name = elem.getAttribute(GSXML.NAME_ATT);
		String module_type = elem.getAttribute(GSXML.TYPE_ATT);
		if (action.equals(GSXML.CONFIG_ACTION_ACTIVATE)) {
		    if (module_type.equals(GSXML.COLLECTION_ELEM)) {
			activateCollection(module_name);
		    } else if (module_type.equals(GSXML.SERVICE_CLASS_ELEM)) {
			activateServiceRack(module_name);
		    } else if (module_type.equals(GSXML.CLUSTER_ELEM)) {
			activateServiceCluster(module_name);
		    } else {
			// cant process the activation
			// send an error
		    }
		} else if (action.equals(GSXML.CONFIG_ACTION_DEACTIVATE)) {
			deactivate(module_type, module_name);
			
		}
	    }
	    
	    return response;
	    
	}
	// if get here something has gone wrong
	System.err.println("MessageRouter: cant process request:");
	System.err.println(converter_.getString(req));
	return null;
       
    }

    protected boolean deactivate(String type, String name) {
	
	if (module_map_.containsKey(name)) {

	    System.out.println("MessageRouter: deactivating "+name);
	    module_map_.remove(name);

	    // also remove the xml bit from description list
	    if (type.equals(GSXML.COLLECTION_ELEM)) {
		Element this_col = GSXML.getNamedElement(collection_list_, GSXML.COLLECTION_ELEM, GSXML.NAME_ATT, name);
		collection_list_.removeChild(this_col);
		return true;
	    } else if (type.equals(GSXML.SERVICE_ELEM)) {
		Element this_service = GSXML.getNamedElement(service_list_, GSXML.SERVICE_ELEM, GSXML.NAME_ATT, name);
		service_list_.removeChild(this_service);
		return true;
	    } else if (type.equals(GSXML.CLUSTER_ELEM)) {
		Element this_cluster = GSXML.getNamedElement(cluster_list_, GSXML.CLUSTER_ELEM, GSXML.NAME_ATT, name);
		cluster_list_.removeChild(this_cluster);
		return true;
	    } else {
		System.err.println("couldn't deactivate coll");
		// couldn't do it
		return false;
	    }
	}
	// else not activated 
	return false;

    }
}


